# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Providers related defines."""

import os
import imp
import time
from os.path import join, dirname, abspath
from gettext import gettext as _

from libmace import util, printf, lsb_release, env
from libmace.util import enforce


NOTICE = """\
WARNING: File was generate by mace utility, do not edit it.
         Change configuration sources and run mace."""

PROVIDERS_DIR = join(abspath(dirname(__file__)), 'providers')
_STOP_TRIES = 10


class BaseProvider(object):
    """All providers need to inherit this class."""

    processed_files = []
    applied_files = []

    def __init__(self):
        self.distro = {}
        if hasattr(self, 'DISTRO_MAP'):
            lsb_id = env.distributor_id.value or lsb_release.distributor_id()
            lsb_rel = env.release.value or lsb_release.release()
            self.distro = getattr(self, 'DISTRO_MAP').get((lsb_id, lsb_rel))
            if self.distro is None:
                self.distro = getattr(self, 'DISTRO_MAP').get((lsb_id, None))
            enforce(self.distro is not None,
                    _('Provider %s is not supported on %s-%s'),
                    self.name, lsb_id, lsb_rel)

    @property
    def comment(self):
        """String that is being used to start comment line in config files."""
        raise NotImplementedError()

    @property
    def name(self):
        """Provider identification string."""
        raise NotImplementedError()

    @property
    def provides(self):
        """List of names that are associated with a provider.

        By default providers are associated only with theirs names.

        """
        return [self.name]

    @property
    def service(self):
        """Shell command to control service daemon.

        For example, /etc/init.d/httd to start and stop HTTP server.

        """
        return self.distro.get('service')

    @property
    def service_command(self):
        """Command to run in shell mode to control service daemon."""
        if self.service:
            if lsb_release.distributor_id() == 'Ubuntu':
                return 'service %s' % self.service
            else:
                return '/etc/init.d/%s' % self.service

    @property
    def confiles(self):
        """Configuration files supported by provider.

        Property contains fake names internally associated with final
        configuration files. If there is only one configuration file
        supported by provider, names don't make sense and even None is fine.

        """
        raise NotImplementedError()

    def is_active(self):
        """Is service running."""
        if not self.service:
            return
        if lsb_release.distributor_id() == 'Ubuntu':
            output = util.call(['status', self.service], close_fds=True)
            if output is not None:
                return output.split()[1].startswith('start')
        output = util.call(self.service_command + ' status',
                shell=True, close_fds=True)
        return output is not None

    def confpath(self, confile=None):
        """Get configuration path relatively to --output directory.

        :param confile:
            one of `confiles` names
        :returns:
            path to the final configuration file associated with `filename`,
            TODO, redesign:
            `None` if confile should be treated as an .ini file,
            `False` custom configl processing

        """
        confpath = self.distro.get('confpath')
        if isinstance(confpath, str):
            return confpath
        elif isinstance(confpath, dict):
            return confpath.get(confile)

    def confpath_header(self, confpath, conftext):
        """Final configuration file header.

        The header that will be added to each final configuration file.

        :param confpath:
            path to the final configuration file
        :param conftext:
            `libmace.files.Conftext` object with full configuration
        :returns:
            list of lines to add

        """
        return []

    def confile_header(self, confile, conftext):
        """Confile header.

        These lines will be added to the final configuration file before each
        confile, even if several confiles go to the same final configuration
        file.

        :param confile:
            one of `confiles` names
        :param conftext:
            `libmace.files.Conftext` object with full configuration
        :returns:
            list of lines to add

        """
        return []

    def footer(self, path, conftext):
        """Final configuration file footer.

        The footer that will be added to each final configuration file.

        :param path:
            path to the final configuration file
        :param conftext:
            `libmace.files.Conftext` object with full configuration
        :returns:
            list of lines to add

        """
        return []

    def confile_footer(self, confile, conftext):
        """Confile footer.

        These lines will be added to the final configuration file after each
        confile, even if several confiles go to the same final configuration
        file.

        :param confile:
            one of `confiles` names
        :param conftext:
            `libmace.files.Conftext` object with full configuration
        :returns:
            list of lines to add

        """
        return []

    def activate(self, force=False):
        """Start service.

        :param force:
            force activating without checking current status
        :returns:
            `True` is service was started

        """
        if self.service and (force or not self.is_active()):
            util.assert_call(self.service_command + ' start',
                    shell=True, close_fds=True)
            return True

    def deactivate(self):
        """Stop service.

        :returns:
            `True` is service was stapped

        """
        if not self.service:
            return

        output = util.call(self.service_command + ' stop',
                shell=True, close_fds=True)

        for a_try in range(_STOP_TRIES):
            if not self.is_active():
                break
            printf.info(_('Wait a second for %s stop (%s)'),
                    self.name, a_try + 1)
            time.sleep(1)
        else:
            raise RuntimeError(_('Cannot stop %s properly') % self.name)

        return output is not None

    def export(self, envars):
        """Export variables that might be used in configuration sources.

        :param envars:
            dictionary to export variables to

        """
        pass

    def convert(self, confile, lines):
        """Convert configuration file.

        :param confile:
            one of `confiles` names that lines belong to
        :param lines:
            configuration file text lines
        :returns:
            function should return new lines

        """
        return lines

    def ensure(self, conftext, tran):
        """Does environment needs to be prepared.

        Will be called after applying new configuration to make sure that
        final environment for consistence and tunning it otherwise.

        :param conftext:
            `libmace.files.Conftext` object with full configuration
        :param tran:
            any changes in environment should happen via this object
            to make it possible to revert made changes on applying fails
        :returns:
            `True` if environment was changed

        """
        pass

    def apply(self, conftext, tran):
        """Apply configuration changes, e.g., restart service.

        :param conftext:
            `libmace.files.Conftext` object with full configuration
        :param tran:
            any changes in environment should happen via this object
            to make it possible to revert made changes on applying fails
        :returns:
            `True` if provider has done something for applying changes, e.g.,
            restarted daemon

        """
        self.deactivate()
        return self.activate(force=True)

    def get_depends(self, conftext):
        """Providers that current one requires.

        :param conftext:
            `libmace.files.Conftext` object with full configuration
        :returns:
            list of provider names

        All depending providers need to be included to the same configuration
        as the current one belongs to. It is only about dependencies between
        providers, not between services they configure, e.g., some provider
        might depend on envars exported by enother one.

        """
        return []

    def get_requires(self, conftext):
        """Packages that provider requires.

        :param conftext:
            `libmace.files.Conftext` object with full configuration
        :returns:
            list of native package names

        """
        return self.distro.get('requires', [])

    def write_config(self, tran, confpath, lines):
        """Add finale configuration files to transation.

        :param tran:
            transaction to add files to
        :param confpath:
            destination for configuration file on final system
        :param lines:
            final configuration text

        """
        if self.comment:
            header = ['%s %s' % (self.comment, i) \
                    for i in NOTICE.split('\n')]
            header.append('')
        else:
            header = []
        # The last NL is imprtant, e.g., iptables-restore needs it
        # for the last COMMIT directive
        config = '\n'.join(header + lines) + '\n'
        tran.ensure_file(confpath, content=config)


def load():
    """Populate list of providers.

    Needs to be called on program start.

    """
    result = []

    for mod_file in os.listdir(PROVIDERS_DIR):
        if not mod_file.endswith('.py'):
            continue
        mod_name = mod_file[:-3]
        fp, pathname, description = imp.find_module(mod_name, [PROVIDERS_DIR])
        try:
            mod = imp.load_module(mod_name, fp, pathname, description)
            provider = mod.Provider()
            result.append(provider)
        except Exception:
            printf.exception(_('Cannot load provider from %s file'),
                    join(PROVIDERS_DIR, mod_file))
        finally:
            if fp:
                fp.close()

    return result
