# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Apache
------

Provider will do some extra work:

- make sure that log directories, mentioned in Mace-level configuration,
  are properly created

Mace configuration:
    :Service:
        ``apache.d``
    :Confiles:
        - ``modules``, line per module in format ``module filename``,
          where ``filename`` is exactly a file name, the prefix path will be
          added depending on the final distribution; add modules only on
          purpose preserving modules enabled on distro level
        - ``conf``, the rest of configuration except vhosts
        - ``vhosts``, convenient confile to collect all vhosts declarations
    :Variables:
        - `HTTP_USER` system user http server uses
        - `HTTP_GROUP` system group http server uses

Fedora final configuration:
    :New:
        - ``/etc/httpd/conf.d/zzz-mace.conf``

Ubuntu final configuration:
    :New:
        - ``/etc/apache2/sites-enabled/zzz-mace.conf``

"""
import re
from os.path import exists, join

from libmace.providers import BaseProvider


_LOG_PARAMETER_RE = re.compile('\s*(ErrorLog|CustomLog)\s+([^\s]+)')
_COMMENT_RE = re.compile('#.*$')


class Provider(BaseProvider):

    DISTRO_MAP = {
            ('Fedora', None): {
                'requires': ['httpd'],
                'service': 'httpd',
                'confpath': '/etc/httpd/conf.d/zzz-mace.conf',
                'module_prefix': 'modules',
                'owner': 'apache',
                },
            ('Ubuntu', None): {
                'requires': ['apache2'],
                'service': 'apache2',
                'confpath': '/etc/apache2/sites-enabled/zzz-mace.conf',
                'module_prefix': '/usr/lib/apache2/modules',
                'owner': 'www-data',
                },
            }

    @property
    def comment(self):
        return '#'

    @property
    def name(self):
        return 'apache'

    @property
    def provides(self):
        return [self.name, 'httpd']

    @property
    def confiles(self):
        return ['modules', 'conf', 'vhosts']

    def export(self, envars):
        envars['HTTP_USER'] = envars['HTTP_GROUP'] = self.distro['owner']

    def convert(self, confile, lines):
        if confile != 'modules':
            return lines

        new_lines = []

        for line in lines:
            line = _COMMENT_RE.sub('', line).strip()
            if line:
                module, filename = line.split()
                filename = join(self.distro['module_prefix'], filename)
                new_lines.append('LoadModule %s %s' % (module, filename))

        return new_lines

    def ensure(self, conftext, tran):
        config_path = tran.path(self.confpath())
        if not exists(config_path):
            return

        for line in file(config_path).readlines():
            match = _LOG_PARAMETER_RE.match(line)
            if match is None:
                continue
            logpath = tran.path(match.group(2))
            tran.ensure_file(logpath, owner=self.distro['owner'],
                    create_missed=True)
