# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
dnsmasq
-------

Provider will change ``/etc/resolv.conf`` to point it to the localhost.
The upstream server needs to be chnaged using ``/etc/resolv.conf.dnsmasq``.

Mace configuration:
    :Service:
        ``dnsmasq.d``
    :Confiles:
        No
    :Variables:
        No

Final configuration for all distributions:
    :Replaces:
        - ``/etc/resolv.conf``
    :New:
        - ``/etc/resolv.conf.dnsmasq``
        - ``/etc/dnsmasq.d/zzz-mace.conf``

"""
import re
from os.path import exists

from libmace.providers import BaseProvider


_COMMENT_RE = re.compile('#.*$')


class Provider(BaseProvider):

    @property
    def comment(self):
        return '#'

    @property
    def name(self):
        return 'dnsmasq'

    @property
    def service(self):
        return 'dnsmasq'

    def get_depends(self, conftext):
        return ['/etc/hosts']

    def get_requires(self, conftext):
        return ['dnsmasq']

    @property
    def confiles(self):
        return [None]

    def confpath(self, confile=None):
        return '/etc/dnsmasq.d/zzz-mace.conf'

    def ensure(self, conftext, tran):
        tran.ensure_dir('/etc')

        conf = tran.path('/etc/resolv.conf')
        conf_dnsmasq = tran.path('/etc/resolv.conf.dnsmasq')
        upstreams = []
        dowstreamed = False
        dowstream_conf = []

        if exists(conf):
            for line in file(conf).readlines():
                parts = _COMMENT_RE.sub('', line).split()
                if len(parts) > 1 and parts[0] == 'nameserver':
                    if parts[1] in ['127.0.0.1', 'localhost']:
                        upstreams = []
                        dowstreamed = True
                    elif not dowstreamed:
                        upstreams.append(parts[1])
                else:
                    dowstream_conf.append(line.strip())

        if not dowstreamed:
            dowstream_conf.append('nameserver 127.0.0.1')
            tran.ensure_file(conf, content='\n'.join(dowstream_conf))

        if not exists(conf_dnsmasq) or upstreams:
            if not upstreams:
                upstreams = ['8.8.8.8', '8.8.4.4']
            tran.ensure_file(conf_dnsmasq, content='\n'.join(
                    ['nameserver %s' % i for i in upstreams]))

    def footer(self, path, conftext):
        return ['resolv-file=/etc/resolv.conf.dnsmasq']
