# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Files and directories
---------------------

Configure arbitrary files and directories.

This provider uses artificial Mace-level configuration syntax in ini-files
format. Each ini section name is a path to configure.

List of supported options:
    :content:
        if specified, file will be processed and directory otherwise
    :mode:
        file mode; default mode if ommited
    :user:
        owner name or uid; root if ommited
    :group:
        owner group or gid; root if ommited
    :exec:
        call shell command if file was changed

Mace configuration:
    :Service:
        ``files.d``
    :Confiles:
        No
    :Variables:
        No

Final configuration for all distributions:
    Depends on mace configuration.

"""
from os.path import exists
from gettext import gettext as _

from libmace import printf, util, env
from libmace.providers import BaseProvider
from libmace.util import enforce


class Provider(BaseProvider):

    DISTRO_MAP = {
            ('Fedora', None): {
                'admin_group': 'wheel',
                },
            ('Ubuntu', None): {
                'admin_group': 'adm',
                },
            }

    @property
    def comment(self):
        return None

    @property
    def name(self):
        return 'files'

    @property
    def confiles(self):
        return [None]

    def confpath(self, confile=None):
        return None

    def export(self, envars):
        envars['ADMIN_GROUP'] = self.distro['admin_group']

    def apply(self, conftext, tran):
        applied = False

        for path in tran.config.sections():
            mode = tran.config[path, 'mode']
            if mode:
                enforce(mode.isdigit, _('File mode should be a digit in ' \
                        'configuration files for files.d provider'))
                mode = eval(mode)

            user = tran.config[path, 'owner'] or 'root'
            group = tran.config[path, 'group'] or user
            owner = (user, group)

            content = tran.config[path, 'content']
            if content:
                stale = tran.ensure_file(path, content=content, owner=owner,
                        mode=mode, abs_path=True)
            else:
                stale = tran.ensure_dir(path, owner=owner, mode=mode,
                        abs_path=True)

            BaseProvider.processed_files.append(path)

            exec_cmd = tran.config[path, 'exec']
            if stale and exec_cmd:
                try:
                    util.assert_call(exec_cmd, shell=True, close_fds=True)
                except Exception:
                    printf.exception(_('Cannot apply %s changes'), path)
                    if not env.no_rollback.value:
                        printf.info(_('Rollback %s'), self.name)
                        tran.rollback()
                        stale = False
                        if exists(path):
                            try:
                                util.assert_call(exec_cmd,
                                        shell=True, close_fds=True)
                                printf.info(_('Reaplied %s with old content'),
                                        path)
                            except Exception:
                                printf.exception(
                                        _('Cannot reaply %s with old content'),
                                        path)

            if stale:
                BaseProvider.applied_files.append(path)

            applied = applied or stale

        return applied
