# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
IPTables
--------

All configuration content should be in ``iptables-save`` format.
But ``COMMIT`` directive **should not** be used, it will be added automatically
by Mace.

Mace configuration:
    :Service:
        - ``iptables.d``
    :Confiles:
        - ``mangle`` for mangle table rules
        - ``nat`` for nat table rules
        - ``filter`` for filter table rules
    :Variables:
        No

Fedora final configuration:
    :Replaces:
        - ``/etc/sysconfig/iptables``

Ubuntu final configuration:
    :New:
        - ``/etc/init.d/iptables``
        - ``/etc/iptables.rules``

"""
from libmace.providers import BaseProvider


class Provider(BaseProvider):

    DISTRO_MAP = {
            ('Fedora', None): {
                'requires': ['iptables'],
                'confpath': '/etc/sysconfig/iptables',
                },
            ('Ubuntu', None): {
                'confpath': '/etc/iptables.rules',
                },
            }

    @property
    def comment(self):
        return '#'

    @property
    def name(self):
        return 'iptables'

    @property
    def service(self):
        return 'iptables'

    @property
    def confiles(self):
        return ['mangle', 'nat', 'filter']

    def confile_header(self, confile, conftext):
        return ['*%s' % confile]

    def confile_footer(self, confile, conftext):
        return ['COMMIT']

    def ensure(self, conftext, tran):
        if 'requires' not in self.distro:
            return tran.ensure_file('/etc/init.d/iptables',
                    content=_IPTABLES_D % self.distro, mode=0755)


_IPTABLES_D = """\
#!/bin/sh

RETVAL=0
RULES="%(confpath)s"
RULES_STAMP="/var/run/iptables.stamp"

flush() {
    iptables -F
    iptables -X
    iptables -t nat -F
    iptables -t nat -X
    iptables -t mangle -F
    iptables -t mangle -X
    iptables -P INPUT ACCEPT
    iptables -P FORWARD ACCEPT
    iptables -P OUTPUT ACCEPT
    rm -f $RULES_STAMP
}

stamp() {
    iptables-save | grep -v ^# | sed 's/\[[0-9]\+:[0-9]\+\]/\[0:0\]/'
}

case "$1" in
    start)
        echo -n "Starting iptables..."
        flush
        test -e $RULES && iptables-restore < $RULES
        RETVAL=$?
        if [ $RETVAL -eq 0 ]; then
            stamp > $RULES_STAMP
            echo "ok"
        else
            echo "failed"
        fi
        ;;
    stop)
        echo -n "Stopping iptables..."
        flush
        echo "ok"
        ;;
    status)
        test -e $RULES_STAMP && stamp | diff - $RULES_STAMP >/dev/null
        RETVAL=$?
        ;;
    reload|restart)
        $0 stop
        $0 start
        RETVAL=$?
        ;;
    *)
        echo "Usage: iptables {start|stop|status|restart}"
        exit 1
esac

exit $RETVAL
"""
