# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Munin node
----------

Provider will do some extra work:

- escape IPs in allow directive in munin-node.conf
- resolve hostnames in allow directive in munin-node.conf
- make sure that only mace level plugins exist

Mace configuration:
    :Service:
        ``munin-node.d``
    :Confiles:
        - ``conf`` configuration
        - ``plugins`` node plugin names per line; for wildcard plugins,
          the suffix needs to be space separated
    :Variables:
        No

Fedora final configuration:
    :Replaces:
        - ``/etc/munin/munin-node.conf``
    :Removes:
        - ``/etc/munin/plugins/``
    :New:
        - ``/etc/munin/plugins/``

"""
import os
import re
import socket
from os.path import join
from gettext import gettext as _

from libmace.providers import BaseProvider
from libmace.util import enforce


_PLUGINS_DIR = '/etc/munin/plugins'
_IP_RE = re.compile('([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})')
_HOST_RE = re.compile('([\w.]+)')


class Provider(BaseProvider):

    @property
    def comment(self):
        return '#'

    @property
    def name(self):
        return 'munin-node'

    @property
    def service(self):
        return 'munin-node'

    def get_requires(self, conftext):
        return ['munin-node']

    @property
    def confiles(self):
        return ['conf', 'plugins']

    def confpath(self, confile=None):
        if confile == 'conf':
            return '/etc/munin/munin-node.conf'
        else:
            return False

    def write_config(self, tran, confpath, lines):
        if not confpath:
            self._ensure_plugins(tran, lines)
        else:
            BaseProvider.write_config(self, tran, confpath, lines)

    def convert(self, confile, lines):
        result = []
        for line in lines:
            line = line.strip()
            if line.startswith('allow'):
                host = line.split(None, 1)[-1]
                match = _IP_RE.search(host)
                if match is not None:
                    host = _IP_RE.sub('\.'.join(match.groups()), host)
                else:
                    match = _HOST_RE.search(host)
                    if match is not None:
                        try:
                            host_ip = socket.gethostbyaddr(
                                    match.group(1))[-1][0]
                        except socket.error:
                            pass
                        else:
                            host = _HOST_RE.sub(
                                    host_ip.replace('.', '\.'), host)
                line = 'allow %s' % host
            result.append(line)
        return result

    def plugin_path(self, name):
        return '/usr/share/munin/plugins/%s' % name

    def _ensure_plugins(self, tran, lines):
        files = {}

        for line in lines:
            parts = line.split()
            enforce(len(parts) <= 2,
                    _('Plugins line in munin-node.d/plugins should be ' \
                            'in form of "<plugin>[ <wildcard-suffix>]"'))
            if len(parts) == 1:
                files[line] = self.plugin_path(line)
            else:
                plugin, suffix = parts
                files[plugin + suffix] = self.plugin_path(plugin)

        for i in set(os.listdir(_PLUGINS_DIR)) - set(files.keys()):
            tran.ensure_absence(join(_PLUGINS_DIR, i), abs_path=True)

        for dst_filename, src_path in files.items():
            tran.ensure_symlink(src_path, join(_PLUGINS_DIR, dst_filename))
