# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Squid
-----

Provider will do some extra work to make sure that:

- swap directory is properly created
- log directories are properly created
- squidGuard db files have proper ownership

If configuration template contains ``filter`` confile, squidQuard
will be enabled.

Mace configuration:
    :Service:
        ``squid.d``
    :Confiles:
        - ``access`` only for access rules
        - ``conf`` the rest of configuration
        - ``filter`` squidQuard configuration
    :Variables:
        No

Fedora final configuration:
    :Replaces:
        - ``/etc/squid/squid.conf``
        - ``/etc/squid/squidGuard.conf``

"""
import os
import logging
from os.path import dirname, join, exists
from gettext import gettext as _

from libmace import util, env, lsb_release
from libmace.providers import BaseProvider
from libmace.util import enforce


class Provider(BaseProvider):

    DISTRO_MAP = {
            ('Fedora', None): {
                'requires': ['squid'],
                'guard_requires': ['squidGuard'],
                'service': 'squid',
                'confpath': '/etc/squid',
                'owner': 'squid',
                'bin': 'squid',
                },
            ('Ubuntu', None): {
                'requires': ['squid3'],
                'guard_requires': ['squidguard'],
                'service': 'squid3',
                'confpath': '/etc/squid3',
                'owner': 'proxy',
                'bin': 'squid3',
                },
            }

    @property
    def comment(self):
        return '#'

    @property
    def name(self):
        return 'squid'

    def get_requires(self, conftext):
        result = self.distro['requires']
        if conftext.by_confile('filter'):
            result.extend(self.distro['guard_requires'])
        return result

    @property
    def confiles(self):
        return ['access', 'conf', 'filter']

    def is_active(self):
        if lsb_release.distributor_id() != 'Ubuntu':
            return BaseProvider.is_active(self)
        status = util.call(['/usr/bin/pgrep', '-u', self.distro['owner'],
                '-f', 'squid'])
        return status is not None

    def confpath(self, confile=None):
        if confile == 'filter':
            return join(self.distro['confpath'], 'squidGuard.conf')
        else:
            return join(self.distro['confpath'], 'squid.conf')

    def footer(self, path, conftext):
        if path == self.confpath('conf') and conftext.by_confile('filter'):
            return ['url_rewrite_program /usr/bin/squidGuard -c %s' % \
                        self.confpath('filter'),
                    ]
        else:
            return []

    def ensure(self, conftext, tran):
        owner = self._ensure_conf_dirs(tran)

        # On Fedora, init.d script uses hard coded log dir for its output
        tran.ensure_dir('/var/log/squid')

        if conftext.by_confile('filter'):
            self._ensure_db_files(tran, owner)

    def _ensure_db_files(self, tran, owner):
        dbhome = None
        logdir = None

        for line in file(tran.path(self.confpath('filter'))).readlines():
            if line.startswith('dbhome'):
                dbhome = line.split()[-1]
            elif line.startswith('logdir'):
                logdir = line.split()[-1]
        enforce(dbhome, _('Option "dbhome" should be set in ' \
                'squid/filter confile'))
        enforce(logdir, _('Option "logdir" should be set in ' \
                'squid/filter confile'))

        dbhome_path = tran.path(dbhome)
        enforce(exists(dbhome_path),
                _('Squid "dbhome" %s directory does not exist'), dbhome_path)
        for root, __, filenames in os.walk(dbhome_path):
            for i in filenames:
                if i in ['domains', 'urls']:
                    path = join(root, i)
                    if not exists(path + '.db') or \
                            os.stat(path).st_mtime > \
                            os.stat(path + '.db').st_mtime:
                        util.assert_call(['/usr/bin/squidGuard',
                                '-c', self.confpath('filter'), '-C', path])
                        if exists(path + '.db'):
                            # If this particular file is not mentioned
                            # in config, .db file won't be created
                            tran.invalidate()
                            i += '.db'
                if i.endswith('.db'):
                    tran.ensure_file(join(root, i), mode=0664, owner=owner,
                            abs_path=True)

        tran.ensure_dir(logdir, owner=owner)
        for i in os.listdir(tran.path(logdir)):
            tran.ensure_file(join(logdir, i), owner=owner, mode=0640)

    def _ensure_conf_dirs(self, tran):
        user = self.distro['owner']
        group = self.distro['owner']
        swap_path = '/var/spool/squid'
        access_log = '/var/log/squid/access.log'
        cache_log = '/var/log/squid/cache.log'
        cache_store_log = '/var/log/squid/store.log'

        for line in file(tran.path(self.confpath())).readlines():
            if line.startswith('cache_effective_user'):
                user = line.split()[-1]
            elif line.startswith('cache_effective_group'):
                group = line.split()[-1]
            elif line.startswith('cache_dir'):
                swap_path = line.split()[2]
            elif line.startswith('access_log'):
                access_log = line.split()[1]
            elif line.startswith('cache_log'):
                cache_log = line.split()[1]
            elif line.startswith('cache_store_log'):
                cache_store_log = line.split()[1]

        owner = (user, group)

        if tran.ensure_dir(swap_path, owner=owner):
            if not env.dry_services.value:
                logging.info(_('Create Squid swap'))
                util.assert_call([self.distro['bin'], '-z'])

        for i in [access_log, cache_log, cache_store_log]:
            if i != 'none':
                tran.ensure_dir(dirname(i), owner=owner, mode=0755)

        return owner
